"""
Module to define class for SVG ellipse elements.
"""

import xml.etree.ElementTree as ET
from typing_extensions import Unpack
import osvg.elements.elementbase
import osvg.elements.circlebase
import osvg.elements.ellipse
import osvg.positions
import osvg.float
import osvg.float_math


class CircleParams(osvg.elements.elementbase.SVGElementParams):
    """
    Keyword argument definition for Ellipse class.
    """

    radius: float | osvg.float.Float


class Circle(osvg.elements.circlebase.CircleBase):
    """
    Base class for ellipse elements.
    """

    xml_tag = "circle"
    radius = osvg.float.FloatProperty()

    def __init__(self, **kwargs: Unpack[CircleParams]) -> None:
        super().__init__(**kwargs)
        self.radius = kwargs["radius"]
        self.create_default_connectors()

    @property
    def _plain_etree_element(self) -> ET.Element:
        # pylint: disable=f-string-without-interpolation
        element = super()._plain_etree_element
        element.set("cx", str(self.position.x))
        element.set("cy", str(self.position.y))
        element.set("r", str(self.radius))
        return element

    def angled_position(
        self, angle: osvg.float.Float | float
    ) -> osvg.positions.Position:
        return osvg.positions.Position(
            x=osvg.positions.XonCircle(
                angle=angle,
                radius=self.radius,
            ),
            y=osvg.positions.YonCircle(
                angle=angle,
                radius=self.radius,
            ),
        )



"""
Module to define class for SVG ellipse elements.
"""

import xml.etree.ElementTree as ET
from typing import TypedDict
from collections.abc import Iterable
from typing_extensions import Unpack
import osvg.elements.elementbase
import osvg.positions
import osvg.float
import osvg.float_math


class GlyphRotation(osvg.float.FloatList):
    """
    Text rotate
    """

    name = "glyph_rotation"
    svg_attribute = "rotate"


class TextOrientationParams(TypedDict):
    """
    Additional Keyword argument definitions for Text classes.
    """

    width: float | osvg.float.Float
    width_adjust: str
    alignment_baseline: str
    text_anchor: str
    word_spacing: str
    writing_mode: str
    glyph_rotation: Iterable[float | osvg.float.Float]


TextOrientationParamsInfo = {
    "width_adjust": {
        "default": "spacing",
        "svg_attribute": "textAdjust",
    },
    "alignment_baseline": {
        "default": "auto",
        "svg_attribute": "alignment-baseline",
    },
    "text_anchor": {
        "default": "start",
        "svg_attribute": "text-anchor",
    },
    "word_spacing": {
        "default": "normal",
        "svg_attribute": "word-spacing",
    },
    "writing_mode": {
        "default": "horizontal-tb",
        "svg_attribute": "writing-mode",
    },
}


class _TextBase(osvg.elements.elementbase.SVGElement):
    """
    Base class for Text and TextSpan classes.
    """

    width = osvg.float.FloatProperty()

    def __init__(self, text: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.text = text
        self.text_anchor = None
        self.width = kwargs.get("width", 0)
        for attr, info in TextOrientationParamsInfo.items():
            self.__dict__[attr] = kwargs.get(attr, info["default"])
        self.glyph_rotation = GlyphRotation(values=kwargs.get("glyph_rotation", None))
        self.create_default_connectors()

    def create_default_connectors(self):
        """
        Create Positions as connectors and add them.
        """
        x_shift_left = osvg.float.Float(0)
        x_shift_right = osvg.float.Float(0)
        if self.text_anchor == "start":
            x_shift_right = self.width
        elif self.text_anchor == "middle":
            x_shift_left = self.width / -2
            x_shift_right = self.width / 2
        elif self.text_anchor == "end":
            x_shift_left = -self.width
        self.add_connector(
            position=osvg.positions.XShiftedPosition(
                origin=self.position, x_shift=x_shift_left
            ),
            name="left",
            respect_rotation=True,
        )
        self.add_connector(
            position=osvg.positions.XShiftedPosition(
                origin=self.position, x_shift=x_shift_right
            ),
            name="right",
            respect_rotation=True,
        )

    @property
    def _plain_etree_element(self) -> ET.Element:
        element = super()._plain_etree_element
        element.text = self.text
        return element

    def _add_orientation_attributes(self, element: ET.Element) -> None:
        """
        Add "orientation" SVG attributes to element
        """
        if str(self.width) != "0":
            element.set("textLenght", str(self.width))
        for attr, info in TextOrientationParamsInfo.items():
            if self.__dict__[attr] != info["default"]:
                element.set(info["svg_attribute"], str(self.__dict__[attr]))
        if not self.glyph_rotation.isdefault:
            element.set(self.glyph_rotation.svg_attribute, str(self.glyph_rotation))


class TextSpanParams(TextOrientationParams):
    """
    Keyword argument definition for TextSpan class.
    """

    dx: float | osvg.float.Float | str = None
    dy: float | osvg.float.Float | str = None
    align_with_parent: bool = False
    style: osvg.style.Style = None
    hyperlink: str = None


class TextSpan(_TextBase):
    """
    SVGP Tspan class. See: https://developer.mozilla.org/en-US/docs/Web/SVG/Element/tspan
    """

    xml_tag = "tspan"
    x = osvg.float.FloatProperty()
    y = osvg.float.FloatProperty()

    def __init__(
        self, parent: "TextSpan | Text", text: str, **kwargs: Unpack[TextSpanParams]
    ) -> None:
        if not isinstance(parent, _TextBase):
            raise TypeError("Parent must be a subclass of Text or TextSpan")
        kwargs["parent"] = parent
        self.align_with_parent = kwargs.get("align_with_parent", False)
        kwargs["width"] = kwargs.get(
            "width", parent.width if self.align_with_parent else 0
        )
        super().__init__(text=text, **kwargs)
        self._dx = None
        self.dx = kwargs.get("dx", None)
        self._dy = None
        self.dy = kwargs.get("dy", None)

    @property
    def dx(self):
        """
        Get dx
        """
        return self._dx

    @dx.setter
    def dx(self, dx: float | osvg.float.Float | str):
        """
        Set dx
        """
        if dx is None:
            self.x = self.parent.position.x if self.align_with_parent else None
            self._dx = None
            return
        self.x = None
        if isinstance(dx, str):
            self._dx = dx
            return
        if not isinstance(dx, osvg.float.Float):
            dx = osvg.float.Float(dx)
        self._dx = osvg.float.Float(dx)

    @property
    def dy(self):
        """
        Get dy
        """
        return self._dy

    @dy.setter
    def dy(self, dy: float | osvg.float.Float | str):
        """
        Set dy
        """
        if dy is None:
            self.y = self.parent.position.y if self.align_with_parent else None
            self._dy = None
            return
        self.y = None
        if isinstance(dy, str):
            self._dy = dy
            return
        if not isinstance(dy, osvg.float.Float):
            dy = osvg.float.Float(dy)
        self._dy = osvg.float.Float(dy)

    @property
    def _plain_etree_element(self) -> ET.Element:
        element = super()._plain_etree_element
        if self.dx is None:
            if self.x:
                element.set("x", str(self.x))
        else:
            element.set("dx", str(self.dx))
        if self.dy is None:
            if self.y:
                element.set("y", str(self.y))
        else:
            element.set("dy", str(self.dy))
        self._add_orientation_attributes(element=element)
        self.add_element_rotation(element=element)
        return element


class TextParams(TextOrientationParams, osvg.elements.elementbase.SVGElementParams):
    """
    Keyword argument definitions for Text classes.
    """


class Text(_TextBase):
    """
    SVG Text class. See: https://developer.mozilla.org/en-US/docs/Web/SVG/Element/text

    Parameters:
        - text: str
        - width ==> "textLength"
        - width_adjust ==> "lengthAdjust"
        - glyph_rotation ==> "rotate"
        - alignment_baseline ==> "alignment-baseline":
                https://developer.mozilla.org/en-US/docs/Web/SVG/Attribute/alignment-baseline
        - text_anchor ==> "text-anchor":
                https://developer.mozilla.org/en-US/docs/Web/SVG/Attribute/text-anchor
        - word_spacing ==> "word-spacing":
                https://developer.mozilla.org/en-US/docs/Web/SVG/Attribute/word-spacing
        - writing_mode ==> "writing-mode":
                https://developer.mozilla.org/en-US/docs/Web/SVG/Attribute/writing-mode
    """

    xml_tag = "text"

    def __init__(self, text: str, **kwargs: Unpack[TextParams]) -> None:
        super().__init__(text=text, **kwargs)

    @property
    def _plain_etree_element(self) -> ET.Element:
        element = super()._plain_etree_element
        element.set("x", str(self.position.x))
        element.set("y", str(self.position.y))
        self._add_orientation_attributes(element=element)
        self.add_element_rotation(element=element)
        return element

    def add_text_span(self, text: str, **kwargs: Unpack[TextSpanParams]) -> None:
        """
        Add a TextSpan object to this Text object
        """
        TextSpan(parent=self, text=text, **kwargs)



"""
Object-oriented SVG Drawing Tool
"""

from .float import (
    Float,
    Sum,
    Sub,
    Mul,
    Div,
    FloorDiv,
    Mod,
    Pow,
)
from .float_math import (
    Cos,
    Sin,
    Cot,
    Tan,
    FunctFloat,
    Min,
    LowestAbs,
    Max,
)
from .elements.circle import Circle
from .elements.elementbase import (
    Height,
    Width,
)
from .elements.ellipse import Ellipse
from .elements.group import Group
from .elements.line import Line
from .elements.line_marker import (
    CircleMarker,
    ArrowMarker,
    LightArrowMarker,
    SkinnyArrowMarker,
)
from .elements.text import Text, TextSpan
from .elements.path import (
    Path,
    PathA,
    PathC,
    Pathc,
    PathH,
    Pathh,
    PathL,
    Pathl,
    PathM,
    Pathm,
    PathQ,
    Pathq,
    PathS,
    Paths,
    PathT,
    Patht,
    PathV,
    Pathv,
    PathZ,
)
from .elements.polyline import Polyline
from .elements.rectangle import Rectangle, SCRectangle
from .elements.svg import SVG, ViewBox
from .positions import (
    Position,
    XShiftedPosition,
    YShiftedPosition,
    PolarShiftedPosition,
    RotatedPosition,
    Distance,
    AngleDegree,
)
from .style import (
    FillColor,
    FillOpacity,
    StrokeDashArray,
    StrokeColor,
    StrokeLineCap,
    StrokeLineJoin,
    StrokeOpacity,
    StrokeWidth,
    Style,
)

__version__ = "2.0.1"
version_info = (2, 0, 1)



"""
Module to define the basic Float classes.
"""

from typing import Union
from collections.abc import Iterable


class Float:
    """
    A Float is an object that has other Float objects as inputs and output.
    With Float you can build a loop-free directed graph for calculating
    dynamically float numbers.
    Float can be used with pyhton built-in arithmetic operations.
    """

    str_round = 3

    def __init__(self, value: Union["Float", float] = None) -> None:
        self._value = None
        self.inputs = []
        self.outputs = []
        self._float_ref = None
        if value is not None:
            self.float_ref = value

    @property
    def float_ref(self) -> "Float":
        """
        Get the reference Float object
        """
        return self._float_ref

    @float_ref.setter
    def float_ref(self, float_ref: "Float"):
        if isinstance(self._float_ref, Float):
            self.remove_input(self._float_ref)
        if isinstance(float_ref, Float):
            self._float_ref = float_ref
            self.add_input(self._float_ref)
        else:
            if hasattr(float_ref, "__float__"):
                self._float_ref = float_ref
            else:
                self._float_ref = float(float_ref)
            self.input_changed()

    @property
    def _get_value(self) -> float:
        if self._float_ref is None:
            raise TypeError("Object not fully initialized")
        return float(self._float_ref)

    @property
    def input_history(self) -> set["Float"]:
        """
        Get all inputs of all inputs recursively
        """
        input_history = set(self.inputs)
        for i in input_history:
            input_history.union(i.input_history)
        return input_history

    def add_input(self, other: "Float") -> None:
        """
        Add other Float object as a input
        """
        if self in other.input_history:
            raise RuntimeError("Input loop detected")
        self.inputs.append(other)
        other.outputs.append(self)
        self.input_changed()

    def remove_input(self, other: "Float") -> None:
        """
        Remove other Float object from inputs
        """
        self.inputs.remove(other)
        other.outputs.remove(self)
        self.input_changed()

    def add_output(self, other: "Float") -> None:
        """
        Add other Float object as a output
        """
        other.add_input(self)

    def remove_output(self, other: "Float") -> None:
        """
        Remove other Float object from outputs
        """
        other.remove_input(self)

    def my_input_changed(self):
        """
        Perform actions on input change
        """

    def input_changed(self):
        """
        Inform myself and all outputs about an input change
        """
        self.my_input_changed()
        for o in self.outputs:
            o.input_changed()

    def __str__(self):
        v = round(float(self), self.str_round)
        if v == 0:
            v = 0
        return f"{v:g}"

    def __float__(self):
        return self._get_value

    def __int__(self):
        return int(float(self))

    def __bool__(self):
        try:
            return bool(self._get_value)
        except TypeError:
            return False

    def __neg__(self):
        return Mul(value1=self, value2=-1)

    def __pos__(self):
        return Mul(value1=self, value2=1)

    @property
    def _i_action_value(self) -> "Float":
        if isinstance(self.float_ref, Float):
            return self.float_ref.float_ref
        return Float(self.float_ref)

    def __add__(self, other):
        return Sum(value1=self, value2=other)

    def __radd__(self, other):
        return Sum(value1=other, value2=self)

    def __sub__(self, other):
        return Sub(value1=self, value2=other)

    def __rsub__(self, other):
        return Sub(value1=other, value2=self)

    def __mul__(self, other):
        return Mul(value1=self, value2=other)

    def __rmul__(self, other):
        return Mul(value1=other, value2=self)

    def __div__(self, other):
        return Div(value1=self, value2=other)

    def __truediv__(self, other):
        return self.__div__(other)

    def __rdiv__(self, other):
        return Div(value1=other, value2=self)

    def __rtruediv__(self, other):
        return self.__rdiv__(other)

    def __floordiv__(self, other):
        return FloorDiv(value1=self, value2=other)

    def __rfloordiv__(self, other):
        return FloorDiv(value1=other, value2=self)

    def __mod__(self, other):
        return Mod(value1=self, value2=other)

    def __rmod__(self, other):
        return Mod(value1=other, value2=self)

    def __pow__(self, other):
        return Pow(value1=self, value2=other)

    def __rpow__(self, other):
        return Pow(value1=other, value2=self)


class FloatList:
    """
    List of Float values
    """

    name = "values"

    def __init__(self, values: Iterable[float | Float] = None) -> None:
        self._values = []
        if values is not None:
            self.values = values

    @property
    def values(self) -> list[Float]:
        """
        Get value list
        """
        return self._values

    @values.setter
    def values(self, values: Iterable[float | Float]) -> None:
        """
        Verify and set values
        """
        if not isinstance(values, list):
            raise TypeError(f"{self.name} must be a list of float values")
        for value in values:
            if not isinstance(value, Float):
                value = Float(value)
            self._values.append(value)

    def __str__(self) -> str:
        if not self.values:
            return ""
        return ",".join([str(value) for value in self.values])

    @property
    def isdefault(self) -> bool:
        """
        Check, if object's value is not set.
        """
        return self.values == []

    def equals(self, other) -> bool:
        """
        Check, if other's float value is equal to self's float value
        """
        for i, v in enumerate(self.values):
            if float(v) != float(other.values[i]):
                return False
        return True


class FloatProperty:
    """
    Property descriptor for Float objects.
    It automatically creates an intermediate Float object,
    which uses the given Float object as input.
    Or, if input is a python native float or int, it creates
    a Float object with that value and use it as the input.
    """

    def __init__(self):
        self.instances = {}

    def __get__(self, instance, owner):
        return self.instances[instance]

    def __set__(self, instance, f):
        if f is not isinstance(f, Float):
            f = Float(f)
        if instance not in self.instances:
            self.instances[instance] = f
        else:
            self.instances[instance].float_ref = f.float_ref


class FloatMath(Float):
    """
    Base class for all "calculated" Float classes
    """

    def __init__(self) -> None:
        super().__init__()
        self._value_cache = None

    def fill_cache(self) -> None:
        """
        Calculate the value for the cache
        """
        raise NotImplementedError

    @property
    def _get_value(self) -> float:
        if self._value_cache is None:
            self.fill_cache()
        return self._value_cache

    def my_input_changed(self):
        self._value_cache = None
        return super().my_input_changed()


class OneValueMath(FloatMath):
    """
    Base class for all arithmetic operations with two Float values
    """

    def __init__(self, value: Float | float) -> None:
        super().__init__()
        self.float_ref = value

    def fill_cache(self) -> None:
        """
        Calculate the value for the cache
        """
        raise NotImplementedError


class TwoValueMath(FloatMath):
    """
    Base class for all arithmetic operations with two Float values
    """

    def __init__(self, value1: Float | float, value2: Float | float) -> None:
        super().__init__()
        self.value1 = value1
        if isinstance(value1, Float):
            self.add_input(value1)
        self.value2 = value2
        if isinstance(value2, Float):
            self.add_input(value2)

    def fill_cache(self) -> None:
        """
        Calculate the value for the cache
        """
        raise NotImplementedError


class Sum(TwoValueMath):
    """
    Represent the sum of two referenced Float objects
    """

    def fill_cache(self) -> None:
        self._value_cache = float(self.value1) + float(self.value2)


class Sub(TwoValueMath):
    """
    Represent the substraction of two referenced Float objects
    """

    def fill_cache(self) -> None:
        self._value_cache = float(self.value1) - float(self.value2)


class Mul(TwoValueMath):
    """
    Represent the multiplication of two referenced Float objects
    """

    def fill_cache(self) -> None:
        self._value_cache = float(self.value1) * float(self.value2)


class Div(TwoValueMath):
    """
    Represent the division of two referenced Float objects
    """

    def fill_cache(self) -> None:
        self._value_cache = float(self.value1) / float(self.value2)


class FloorDiv(TwoValueMath):
    """
    Represent the integer value of a division of two referenced Float objects
    """

    def fill_cache(self) -> None:
        self._value_cache = float(self.value1) // float(self.value2)


class Mod(TwoValueMath):
    """
    Represent the value of a modulo division of two referenced Float objects
    """

    def fill_cache(self) -> None:
        self._value_cache = float(self.value1) % float(self.value2)


class Pow(TwoValueMath):
    """
    Represent the value of the power of two referenced Float objects
    """

    def fill_cache(self) -> None:
        self._value_cache = float(self.value1) ** float(self.value2)



"""
Module to define class for SVG ellipse elements.
"""

import xml.etree.ElementTree as ET
from typing_extensions import Unpack
import osvg.elements.elementbase
import osvg.elements.circlebase
import osvg.positions
import osvg.float
import osvg.float_math


class EllipseParams(osvg.elements.elementbase.SVGElementParams):
    """
    Keyword argument definition for Ellipse class.
    """

    radius_x: float | osvg.float.Float
    radius_y: float | osvg.float.Float


class Ellipse(osvg.elements.circlebase.CircleBase):
    """
    Base class for ellipse elements.
    """

    xml_tag = "ellipse"
    radius_x = osvg.float.FloatProperty()
    radius_y = osvg.float.FloatProperty()

    def __init__(self, **kwargs: Unpack[EllipseParams]) -> None:
        super().__init__(**kwargs)
        self.radius_x = kwargs["radius_x"]
        self.radius_y = kwargs["radius_y"]
        self.create_default_connectors()

    @property
    def _plain_etree_element(self) -> ET.Element:
        element = super()._plain_etree_element
        element.set("cx", str(self.position.x))
        element.set("cy", str(self.position.y))
        element.set("rx", str(self.radius_x))
        element.set("ry", str(self.radius_y))
        self.add_element_rotation(element=element)
        return element

    def angled_position(
        self, angle: osvg.float.Float | float
    ) -> osvg.positions.Position:
        return osvg.positions.Position(
            x=osvg.positions.XonEllipse(
                angle=angle, radius_x=self.radius_x, radius_y=self.radius_y
            ),
            y=osvg.positions.YonEllipse(
                angle=angle, radius_x=self.radius_x, radius_y=self.radius_y
            ),
        )



"""
Module to define Line in SVG.
"""

from collections.abc import Iterable
from typing import TypedDict
from typing_extensions import Unpack
import osvg.elements.polyline
import osvg.elements.group
import osvg.style
import osvg.elements.line_marker
from osvg.positions import PositionProperty


class LineParams(TypedDict):
    """
    Keyword argument definition for Line class.
    """

    parent: osvg.elements.group.Group
    start: osvg.positions.Position | Iterable[float | int]
    end: osvg.positions.Position | Iterable[float | int]
    name: str = None
    style: osvg.style.Style = None
    layer: int = 0


class Line(osvg.elements.polyline.Polyline):
    """
    A Line in SVG from position "start" to position "end".
    """

    xml_tag = "polyline"
    start = PositionProperty()
    end = PositionProperty()

    def __init__(self, **kwargs: Unpack[LineParams]) -> None:
        self.start = kwargs.pop("start")
        self.end = kwargs.pop("end")
        kwargs["positions"] = [self.start, self.end]
        super().__init__(**kwargs)

    def add_start_marker(
        self, marker_class: osvg.elements.line_marker.Marker, **kwargs
    ) -> None:
        """
        Add a start marker to the line
        """
        angle = osvg.positions.AngleDegree(a=self.end, b=self.start)
        self.start_marker = marker_class(
            parent=self, position=self.start, angle=angle, **kwargs
        )

    def add_end_marker(
        self, marker_class: osvg.elements.line_marker.Marker, **kwargs
    ) -> None:
        """
        Add a end marker to the line
        """
        angle = osvg.positions.AngleDegree(a=self.start, b=self.end)
        self.end_marker = marker_class(
            parent=self, position=self.end, angle=angle, **kwargs
        )



"""
Module to define a Rectangle in SVG.
"""

import xml.etree.ElementTree as ET
from typing_extensions import Unpack
import osvg.elements.elementbase
import osvg.elements.rectanglebase
import osvg.style


class RectangleParams(osvg.elements.rectanglebase.RectangleBaseParams):
    """
    Keyword argument definition for RectangleBase class.
    """

    rx: float | osvg.float.Float = None
    ry: float | osvg.float.Float = None


class Rectangle(osvg.elements.rectanglebase.RectangleBase):
    """
    A rectangle in SVG - See: https://developer.mozilla.org/en-US/docs/Web/SVG/Element/rect
    """

    xml_tag = "rect"
    rx = osvg.float.FloatProperty()
    ry = osvg.float.FloatProperty()

    def __init__(self, **kwargs: Unpack[RectangleParams]) -> None:
        self.rx = kwargs.get("rx", None)
        self.ry = kwargs.get("ry", None)
        super().__init__(**kwargs)

    @property
    def _plain_etree_element(self) -> ET.Element:
        # pylint: disable=f-string-without-interpolation
        element = super()._plain_etree_element
        element.set(
            "x", f"{round(float(self.position.x) - float(self.width) / 2, 3):g}"
        )
        element.set(
            "y", f"{round(float(self.position.y) - float(self.height) / 2, 3):g}"
        )
        element.set("width", str(self.width))
        element.set("height", str(self.height))
        if self.rx:
            element.set("rx", str(self.rx))
        if self.ry:
            element.set("ry", str(self.ry))
        self.add_element_rotation(element=element)
        return element


class SCRectangleParams(osvg.elements.rectanglebase.RectangleBaseParams):
    """
    Keyword argument definition for RectangleBase class.
    """

    percentage: float | osvg.float.Float = 5


class SCRectangle(Rectangle):
    """
    A Smooth Cornered Rectangle

    Parameter:
      percentage: Percent of the lower value of width or height.
    """

    percentage = osvg.float.FloatProperty()

    def __init__(self, **kwargs: Unpack[SCRectangleParams]) -> None:
        super().__init__(**kwargs)
        self.percentage = kwargs.get("percentage", 5)
        self.rx = osvg.float_math.PercentOf(
            osvg.float_math.Min(self.width, self.height), self.percentage
        )



"""
Module to define position objects.
"""

# pylint: disable=missing-function-docstring

import math
from collections.abc import Iterable
import osvg.float
import osvg.float_math
import osvg.helper


class Position:
    """
    Class to define a point within in the two-dimensional drawing.
    'x' is the floating point value on the x axis with
    x=0 as the most left point and x>0 is moving to the right.
    'y' is the floating point value on the y axis with
    y=0 as the most upper point and y>0 is moving down.

    A PositionObject is a linkable object.
    """

    # pylint: disable=too-few-public-methods

    str_round = 3
    x = osvg.float.FloatProperty()
    y = osvg.float.FloatProperty()

    def __init__(
        self,
        x: osvg.float.Float | float = 0.0,
        y: osvg.float.Float | float = 0.0,
    ) -> None:
        self.x = x
        self.y = y

    def __str__(self):
        """
        Get the SVG string
        """
        return f"{self.x},{self.y}"

    @staticmethod
    def __clean_other(other):
        if not isinstance(other, Position):
            if isinstance(other, Iterable):
                return Position(*other)
            return Position(other, other)
        return other

    def __add__(self, other):
        other = self.__clean_other(other)
        return Position(x=self.x + other.x, y=self.y + other.y)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        other = self.__clean_other(other)
        return Position(x=self.x - other.x, y=self.y - other.y)

    def __rsub__(self, other):
        other = self.__clean_other(other)
        return Position(x=other.x - self.x, y=other.y - self.y)

    def __mul__(self, other):
        other = self.__clean_other(other)
        return Position(x=self.x * other.x, y=self.y * other.y)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __div__(self, other):
        other = self.__clean_other(other)
        return Position(x=self.x / other.x, y=self.y / other.y)

    def __truediv__(self, other):
        return self.__div__(other)

    def __rdiv__(self, other):
        other = self.__clean_other(other)
        return Position(x=other.x / self.x, y=other.y / self.y)

    def __rtruediv__(self, other):
        return self.__div__(other)

    def __floordiv__(self, other):
        other = self.__clean_other(other)
        return Position(x=self.x // other.x, y=self.y // other.y)

    def __rfloordiv__(self, other):
        other = self.__clean_other(other)
        return Position(x=other.x // self.x, y=other.y // self.y)

    def __mod__(self, other):
        other = self.__clean_other(other)
        return Position(x=self.x % other.x, y=self.y % other.y)

    def __rmod__(self, other):
        other = self.__clean_other(other)
        return Position(x=other.x % self.x, y=other.y % self.y)

    def __pow__(self, other):
        other = self.__clean_other(other)
        return Position(x=self.x**other.x, y=self.y**other.y)

    def __rpow__(self, other):
        other = self.__clean_other(other)
        return Position(x=other.x**self.x, y=other.y**self.y)


class PositionProperty:
    """
    Property descriptor for Position objects
    """

    def __init__(self):
        self.instances = {}

    def __get__(self, instance, owner):
        return self.instances[instance]

    def __set__(
        self, instance, p: Position | Iterable[osvg.float.Float | float] = None
    ):
        if p is None:
            x = 0
            y = 0
        elif issubclass(type(p), Position):
            x = p.x
            y = p.y
        elif isinstance(p, Iterable):
            x = p[0]
            y = p[1]
        else:
            raise TypeError("Position object required")
        if instance not in self.instances:
            self.instances[instance] = Position()
        self.instances[instance].x = x
        self.instances[instance].y = y


class XShiftedPosition(Position):
    # pylint: disable=too-few-public-methods
    """
    A relative position object where the x coordinate is shifted
    by the given float or to the x value of the given position.
    """

    def __init__(
        self,
        origin: Position,
        x_shift: Position | osvg.float.Float | float = 0,
    ) -> None:
        if not isinstance(origin, Position):
            if isinstance(origin, Iterable):
                origin = Position(*origin)
            else:
                origin = Position(origin, origin)
        if isinstance(x_shift, Position):
            x = x_shift.x
        elif isinstance(x_shift, Iterable):
            x = Position(*x_shift).x
        else:
            x = origin.x + x_shift
        super().__init__(x=x, y=origin.y)


class YShiftedPosition(Position):
    # pylint: disable=too-few-public-methods
    """
    A relative position object where the y coordinate is shifted
    by the given float or to the y value of the given position.
    """

    def __init__(
        self,
        origin: Position,
        y_shift: Position | osvg.float.Float | float = 0,
    ) -> None:
        if not isinstance(origin, Position):
            if isinstance(origin, Iterable):
                origin = Position(*origin)
            else:
                origin = Position(origin, origin)
        if isinstance(y_shift, Position):
            y = y_shift.y
        elif isinstance(y_shift, Iterable):
            y = Position(*y_shift).y
        else:
            y = origin.y + y_shift
        super().__init__(x=origin.x, y=y)


class Distance(osvg.float_math.CalculatedKwargs):
    """
    Class to provide a calculated Float object representing the
    distance between two Positions
    """

    def __init__(self, a: Position, b: Position) -> None:
        super().__init__(ax=a.x, ay=a.y, bx=b.x, by=b.y)

    def fill_cache(self) -> None:
        # pylint: disable=no-member
        self._value_cache = math.dist(
            [float(self.ax), -float(self.ay)],
            [float(self.bx), -float(self.by)],
        )


class AngleDegree(osvg.float_math.CalculatedKwargs):
    """
    Class to provide a calculated Float object representing the
    angle from "a" to another position "b" in radians
    """

    def __init__(self, a: Position, b: Position) -> None:
        super().__init__(ax=a.x, ay=a.y, bx=b.x, by=b.y)

    def fill_cache(self) -> None:
        # pylint: disable=no-member
        dx = float(self.bx) - float(self.ax)
        dy = float(self.ay) - float(self.by)  # due to the inverted y-axis
        self._value_cache = osvg.helper.degrees(math.atan2(dy, dx))


class XonCircle(osvg.float_math.CalculatedKwargs):
    """
    Class to provide a calculated Float object representing the
    x value of a position on a circle.
    Note: angle=0 is twelve o'clock and it goes clockwise.
    """

    def __init__(self, angle: osvg.float.Float, radius: osvg.float.Float) -> None:
        super().__init__(angle=angle, radius=radius)

    def fill_cache(self) -> float:
        # pylint: disable=no-member
        self._value_cache = math.cos(osvg.helper.radians(float(self.angle))) * float(
            self.radius
        )


class YonCircle(osvg.float_math.CalculatedKwargs):
    """
    Class to provide a calculated osvg.float.Float object representing the
    y value of a position on a circle.
    Note: angle=0 is twelve o'clock and it goes clockwise.
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, angle: osvg.float.Float, radius: osvg.float.Float) -> None:
        super().__init__(angle=angle, radius=radius)

    def fill_cache(self) -> float:
        # pylint: disable=no-member
        self._value_cache = math.sin(osvg.helper.radians(float(self.angle))) * -float(
            self.radius
        )


class XonEllipse(osvg.float_math.CalculatedKwargs):
    """
    Class to provide a calculated Float object representing the
    x value of a position on a ellipse.
    Note: angle=0 is twelve o'clock and it goes clockwise.
    """

    def __init__(
        self,
        angle: osvg.float.Float,
        radius_x: osvg.float.Float,
        radius_y: osvg.float.Float,
    ) -> None:
        super().__init__(angle=angle, radius_x=radius_x, radius_y=radius_y)

    def fill_cache(self) -> float:
        # pylint: disable=no-member
        self._value_cache = abs(
            float(self.radius_x)
            * float(self.radius_y)
            / math.sqrt(
                float(self.radius_y) ** 2
                + float(self.radius_x) ** 2
                * math.tan(osvg.helper.radians(float(self.angle))) ** 2
            )
        )
        if 90 < float(self.angle) % 360 < 270:
            self._value_cache = -1 * self._value_cache


class YonEllipse(osvg.float_math.CalculatedKwargs):
    """
    Class to provide a calculated Float object representing the
    y value of a position on a ellipse.
    Note: angle=0 is twelve o'clock and it goes clockwise.
    """

    def __init__(
        self,
        angle: osvg.float.Float,
        radius_x: osvg.float.Float,
        radius_y: osvg.float.Float,
    ) -> None:
        super().__init__(angle=angle, radius_x=radius_x, radius_y=radius_y)

    def fill_cache(self) -> float:
        # pylint: disable=no-member
        tan = math.tan(osvg.helper.radians(float(self.angle)))
        if tan == 0:
            self._value_cache = 0.0
        else:
            self._value_cache = abs(
                float(self.radius_x)
                * float(self.radius_y)
                / math.sqrt(
                    float(self.radius_x) ** 2
                    + float(self.radius_y) ** 2
                    / math.tan(osvg.helper.radians(float(self.angle))) ** 2
                )
            )
            if float(self.angle) % 360 > 180:
                self._value_cache = -1 * self._value_cache


class RotatedPosition(Position):
    """
    Class to define a point rotated around another point by an angle.
    'reference' is the Position which should be shifted.
    'center' is the center of the circle with 'parent' on the cirle track.

    Note:
    - angle goes clockwise
    - angle is delta to current angle 'reference' -> 'parent'
    """

    # pylint: disable=too-few-public-methods

    x = osvg.float.FloatProperty()
    y = osvg.float.FloatProperty()
    origin = PositionProperty()
    center = PositionProperty()
    angle = osvg.float.FloatProperty()

    def __init__(
        self,
        origin: Position,
        center: Position,
        angle: osvg.float.Float | float = 0,
    ) -> None:
        super().__init__()
        self.origin = origin
        self.center = center
        self.angle = angle

        distance = Distance(self.center, self.origin)
        new_angle = self.angle + AngleDegree(self.center, self.origin)

        # Calc x
        new_shift_x = XonCircle(angle=new_angle, radius=distance)
        self.x = self.center.x + new_shift_x

        # Calc y
        new_shift_y = YonCircle(angle=new_angle, radius=distance)
        self.y = self.center.y + new_shift_y


class PolarShiftedPosition(Position):
    """
    Class to define a position which is described by a polar coordinate
    with the reference position as center.
    """

    # pylint: disable=too-few-public-methods

    x = osvg.float.FloatProperty()
    y = osvg.float.FloatProperty()
    origin = PositionProperty()
    angle = osvg.float.FloatProperty()
    distance = osvg.float.FloatProperty()
    x_shift = osvg.float.FloatProperty()
    y_shift = osvg.float.FloatProperty()

    def __init__(
        self,
        origin: Position,
        angle: osvg.float.Float | float = 0,
        distance: osvg.float.Float | float = 0,
    ) -> None:
        super().__init__()
        self.origin = origin
        self.angle = angle
        self.distance = distance

        x_shift = osvg.float_math.Cos(self.angle) * self.distance
        self.x = origin.x + x_shift
        y_shift = -osvg.float_math.Sin(self.angle) * self.distance
        self.y = origin.y + y_shift


class RayCrossCirclePosition(Position):
    """
    Class to define a point at which a ray from
    circle center to a position crosses the circle.
    """

    # pylint: disable=too-few-public-methods

    x = osvg.float.FloatProperty()
    y = osvg.float.FloatProperty()
    center = PositionProperty()
    position = PositionProperty()
    radius = osvg.float.FloatProperty()
    x_shift = osvg.float.FloatProperty()
    y_shift = osvg.float.FloatProperty()

    def __init__(
        self,
        center: Position,
        position: Position,
        radius: osvg.float.Float | float,
    ) -> None:
        super().__init__()
        self.center = center
        self.position = position
        self.radius = radius

        angle = AngleDegree(a=self.center, b=self.position)
        self.x = center.x + XonCircle(angle=angle, radius=self.radius)
        self.y = center.y + YonCircle(angle=angle, radius=self.radius)



"""
Module to define shared methods for rectangle-like classes.
"""

from typing import Literal
from typing_extensions import Unpack
import osvg.elements.elementbase
import osvg.positions
import osvg.float
import osvg.float_math


class RectangleBaseParams(osvg.elements.elementbase.SVGElementParams):
    """
    Keyword argument definition for RectangleBase class.
    """

    width: float | osvg.float.Float = 0
    height: float | osvg.float.Float = 0


class RectangleBase(osvg.elements.elementbase.SVGElement):
    # pylint: disable=no-member
    """
    Base class for rectangle-like Elements.
    """
    width = osvg.float.FloatProperty()
    height = osvg.float.FloatProperty()

    def __init__(self, **kwargs: Unpack[RectangleBaseParams]) -> None:
        super().__init__(**kwargs)
        self.width = kwargs.get("width", 0)
        self.height = kwargs.get("height", 0)
        self.create_default_connectors()

    def _connector_top_left(self):
        return self.position - osvg.Position(self.width, self.height) / 2

    def _connector_top_center(self):
        return osvg.positions.YShiftedPosition(
            origin=self.position,
            y_shift=self.height / -2,
        )

    def _connector_top_right(self):
        return self.position + osvg.Position(self.width / 2, self.height / -2)

    def _connector_center_left(self):
        return osvg.positions.XShiftedPosition(
            origin=self.position,
            x_shift=self.width / -2,
        )

    def _connector_center(self):
        return osvg.positions.Position(self.position.x, self.position.y)

    def _connector_center_right(self):
        return osvg.positions.XShiftedPosition(
            origin=self.position,
            x_shift=self.width / 2,
        )

    def _connector_bottom_left(self):
        return self.position + osvg.Position(self.width / -2, self.height / 2)

    def _connector_bottom_center(self):
        return osvg.positions.YShiftedPosition(
            origin=self.position,
            y_shift=self.height / 2,
        )

    def _connector_bottom_right(self):
        return self.position + osvg.Position(self.width, self.height) / 2

    def create_default_connectors(self) -> None:
        """
        Create Positions as connectors and add them.
        """
        self.add_connector(
            self._connector_top_left(), name="top-left", respect_rotation=True
        )
        self.add_connector(
            self._connector_top_center(), name="top-center", respect_rotation=True
        )
        self.add_connector(
            self._connector_top_right(), name="top-right", respect_rotation=True
        )
        self.add_connector(
            self._connector_center_left(), name="center-left", respect_rotation=True
        )
        self.add_connector(self._connector_center(), "center")
        self.add_connector(
            self._connector_center_right(), name="center-right", respect_rotation=True
        )
        self.add_connector(
            self._connector_bottom_left(), name="bottom-left", respect_rotation=True
        )
        self.add_connector(
            self._connector_bottom_center(), name="bottom-center", respect_rotation=True
        )
        self.add_connector(
            self._connector_bottom_right(), name="bottom-right", respect_rotation=True
        )

    def add_connector_at_angle(self, angle: float | osvg.float.Float, name: str):
        """
        Add a connector at element's border at the angle
        in relation to the center.
        """
        # pylint: disable=too-many-return-statements,too-many-branches
        if name in self.connectors:
            raise ValueError(f"Connector with name {name} already exists")
        if isinstance(angle, osvg.float.Float):
            angle = float(angle)
        n_angle = angle % 360
        if n_angle == 0:
            self.connectors[name] = self.connectors["center-right"]
        else:
            if n_angle > 180:
                half_height = self.height / -2
            else:
                half_height = self.height / 2
            if 90 < n_angle < 270:
                half_width = self.width / -2
            else:
                half_width = self.width / 2
            cot_angle = -osvg.float_math.Cot(angle)
            x_shift = osvg.float_math.LowestAbs(
                half_width,
                cot_angle * half_height,
            )
            tan_angle = -osvg.float_math.Tan(angle)
            y_shift = osvg.float_math.LowestAbs(
                half_height,
                tan_angle * half_width,
            )
            self.add_connector(
                position=self.position + (x_shift, y_shift),
                name=name,
                respect_rotation=True,
            )
        return self.connectors[name]

    def add_side_connector(
        self,
        name: str,
        side: Literal["top", "right", "bottom", "left"],
        percentage: float | osvg.float.Float = 50,
    ):
        """
        Add a connector at element's border on given side at the percent of the side's length.
        Percent will be modulo of 100. (-10 => 90; 110 => 10)
        Percent at top/bottom: 0 is left, 100 is right
        Percent at left/right: 0 is bottom, 100 is top
        """
        # pylint: disable=too-many-return-statements,too-many-branches
        if side in ["top", "bottom"]:
            x_shift = (
                osvg.float_math.PercentOf(
                    reference=self.width, percentage=percentage, modulo=True
                )
                - self.width / 2
            )  # Percent of width minus half width
            y_shift = self.height / 2
            if side == "top":
                y_shift = -y_shift
        elif side in ["left", "right"]:
            x_shift = self.width / 2
            if side == "left":
                x_shift = -x_shift
            y_shift = (
                osvg.float_math.PercentOf(
                    reference=self.height, percentage=percentage, modulo=True
                )
                - self.height / 2
            )  # Percent of heigth minus half heigth
        return self.add_connector(
            position=self.position + (x_shift, y_shift),
            name=name,
            respect_rotation=True,
        )



"""
Module to define class for SVG ellipse elements.
"""

# pylint: disable=too-few-public-methods

from xml.etree import ElementTree as ET
from typing import TypedDict
from typing_extensions import Unpack
import osvg.elements.elementbase
import osvg.positions
import osvg.float
import osvg.float_math


class _PathElement:
    """
    Base class for all Path elements
    """

    tag = ""
    needs_previous_path_element = False
    previous_path_element_sub_classes = []

    def verify_previous_path_element(self, element: "_PathElement") -> None:
        """
        Checks if given (previous) path element if the one which is needed
        """
        if not issubclass(type(element), tuple(self.previous_path_element_sub_classes)):
            class_names = " or ".join(
                [x.__name__ for x in self.previous_path_element_sub_classes]
            )
            raise RuntimeError(
                f"{self.__class__.__name__} need sub-class of "
                + f"{class_names} as previous path element"
            )


class _SingleFloatElement(_PathElement):
    """
    Base class for Path elements which uses a position
    """

    ref = osvg.float.FloatProperty()

    def __init__(self, ref: float | osvg.float.Float) -> None:
        self.ref = ref

    def __str__(self) -> str:
        return f"{self.tag} {str(self.ref)}"


class _AbsolutePositionElement(_PathElement):
    """
    Base class for Path elements which uses a position object
    """

    def __init__(self, position: osvg.positions.Position) -> None:
        self.position = position

    def __str__(self) -> str:
        return f"{self.tag} {str(self.position)}"


class _TwoAbsolutePositionElement(_AbsolutePositionElement):
    """
    Base class for Path elements which uses two position objects
    """

    def __init__(
        self, position1: osvg.positions.Position, position2: osvg.positions.Position
    ) -> None:
        super().__init__(position=position1)
        self.position2 = position2

    def __str__(self) -> str:
        return f"{super().__str__()} {str(self.position2)}"


class _ThreeAbsolutePositionElement(_TwoAbsolutePositionElement):
    """
    Base class for Path elements which uses three position objects
    """

    def __init__(
        self,
        position1: osvg.positions.Position,
        position2: osvg.positions.Position,
        position3: osvg.positions.Position,
    ) -> None:
        super().__init__(position1=position1, position2=position2)
        self.position3 = position3

    def __str__(self) -> str:
        return f"{super().__str__()} {str(self.position3)}"


class _RelativePositionElement(_PathElement):
    """
    Base class for Path elements which uses a position shifted by
    float values x1 and y1.
    """

    x = osvg.float.FloatProperty()
    y = osvg.float.FloatProperty()

    def __init__(
        self, x: float | osvg.float.Float, y: float | osvg.float.Float
    ) -> None:
        self.x = x
        self.y = y

    def __str__(self) -> str:
        return f"{self.tag} {str(self.x)},{str(self.y)}"


class _TwoRelativePositionElement(_RelativePositionElement):
    """
    Base class for Path elements which uses two positions shifted by
    float values x1, y1, x2, and y2.
    """

    x2 = osvg.float.FloatProperty()
    y2 = osvg.float.FloatProperty()

    def __init__(
        self,
        x1: float | osvg.float.Float,
        y1: float | osvg.float.Float,
        x2: float | osvg.float.Float,
        y2: float | osvg.float.Float,
    ) -> None:
        super().__init__(x=x1, y=y1)
        self.x2 = x2
        self.y2 = y2

    def __str__(self) -> str:
        return f"{super().__str__()} {str(self.x2)},{str(self.y2)}"


class _ThreeRelativePositionElement(_TwoRelativePositionElement):
    """
    Base class for Path elements which uses three positions shifted by
    float values x1, y1, x2, y2, x3 and y3.
    """

    x3 = osvg.float.FloatProperty()
    y3 = osvg.float.FloatProperty()

    def __init__(
        self,
        x1: float | osvg.float.Float,
        y1: float | osvg.float.Float,
        x2: float | osvg.float.Float,
        y2: float | osvg.float.Float,
        x3: float | osvg.float.Float,
        y3: float | osvg.float.Float,
    ) -> None:
        # pylint: disable=too-many-arguments
        super().__init__(x1=x1, y1=y1, x2=x2, y2=y2)
        self.x3 = x3
        self.y3 = y3

    def __str__(self) -> str:
        return f"{super().__str__()} {str(self.x3)},{str(self.y3)}"


class PathM(_AbsolutePositionElement):
    """
    Class for a Path element which moves the path to a position object
    """

    tag = "M"


class Pathm(_RelativePositionElement):
    """
    Class for a Path element which moves the path to
    a position shifted by x and y float values
    """

    tag = "m"


class PathL(_AbsolutePositionElement):
    """
    Class for a Path element which draws a line to a position object
    """

    tag = "L"


class Pathl(_RelativePositionElement):
    """
    Class for a Path element which draws a line to
    aa position shifted by x and y float values
    """

    tag = "l"


class PathH(_SingleFloatElement):
    """
    Class for a Path element which draws a horizontal line to
    a position (x, current_y)
    """

    tag = "H"

    def __init__(self, x: float | osvg.float.Float) -> None:
        super().__init__(ref=x)


class Pathh(_SingleFloatElement):
    """
    Class for a Path element which draws a horizontal line to
    a position shifted by x float value
    """

    tag = "h"

    def __init__(self, x: float | osvg.float.Float) -> None:
        super().__init__(ref=x)


class PathV(_SingleFloatElement):
    """
    Class for a Path element which draws a vertical line to
    a position (current_x, y)
    """

    tag = "V"

    def __init__(self, y: float | osvg.float.Float) -> None:
        super().__init__(ref=y)


class Pathv(_SingleFloatElement):
    """
    Class for a Path element which draws a horizontal line to
    a position shifted by y float value
    """

    tag = "v"

    def __init__(self, y: float | osvg.float.Float) -> None:
        super().__init__(ref=y)


class PathZ(_PathElement):
    """
    Class for the final 'Z' element which closes the path to
    the starting position
    """

    def __str__(self):
        return "Z"


class CubicBezierCurveElement:
    """
    Meta class to to verify previous path element
    """


class PathC(_ThreeAbsolutePositionElement, CubicBezierCurveElement):
    """
    Class to draw a cubic Bézier curve by using three position objects:
    - postion1: Bézier Point 1
    - postion2: Bézier Point 2
    - postion3: Final Point
    """

    tag = "C"


class Pathc(_ThreeRelativePositionElement, CubicBezierCurveElement):
    """
    Class to draw a cubic Bézier curve by using three relative positions:
    - x1: x Float value of Bézier Point 1
    - y1: y Float value of Bézier Point 1
    - x2: x Float value of Bézier Point 2
    - y2: y Float value of Bézier Point 2
    - x3: x Float value of Final Point
    - y3: y Float value of Final Point
    """

    tag = "c"


class PathS(_TwoAbsolutePositionElement, CubicBezierCurveElement):
    """
    Class to draw a follow-up cubic Bézier curve by using two position objects:
    - postion1: Bézier Point 2
    - postion2: Final Point
    Note: Bézier Point 1 is a reflection of Bézier Point 2 of previous
    cubic Bézier Curve.
    """

    tag = "S"
    needs_previous_path_element = True
    previous_path_element_sub_classes = [CubicBezierCurveElement]


class Paths(_TwoRelativePositionElement, CubicBezierCurveElement):
    """
    Class to draw a follow-up cubic Bézier curve by using two relative positions:
    - x1: x Float value of Bézier Point 2
    - y1: y Float value of Bézier Point 2
    - x2: x Float value of Final Point
    - y2: y Float value of Final Point
    Note: Bézier Point 1 is a reflection of Bézier Point 2 of previous
    cubic Bézier Curve.
    """

    tag = "s"
    needs_previous_path_element = True
    previous_path_element_sub_classes = [CubicBezierCurveElement]


class QuadraticBezierCurveElement:
    """
    Meta class to to verify previous path element
    """


class PathQ(_TwoAbsolutePositionElement, QuadraticBezierCurveElement):
    """
    Class to draw a quadratic Bézier curve by using two position objects:
    - postion1: Bézier Point
    - postion2: Final Point
    """

    tag = "Q"


class Pathq(_TwoRelativePositionElement, QuadraticBezierCurveElement):
    """
    Class to draw a quadratic Bézier curve by using two relative positions:
    - x1: x Float value of Bézier Point
    - y1: y Float value of Bézier Point
    - x2: x Float value of Final Point
    - y2: y Float value of Final Point
    """

    tag = "q"


class PathT(_AbsolutePositionElement, QuadraticBezierCurveElement):
    """
    Class to draw a follow-up quadratic Bézier curve by using a position object:
    - postion: Final Point
    Note: Bézier Point is a reflection of Bezier Point of previous
    quadratic Bézier Curve.
    """

    tag = "T"
    needs_previous_path_element = True
    previous_path_element_sub_classes = [QuadraticBezierCurveElement]


class Patht(_RelativePositionElement, QuadraticBezierCurveElement):
    """
    Class to draw a follow-up quadratic Bézier curve by using a relative position:
    - x: x Float value of Final Point
    - y: y Float value of Final Point
    Note: Bézier Point is a reflection of Bezier Point of previous
    quadratic Bézier Curve.
    """

    tag = "t"
    needs_previous_path_element = True
    previous_path_element_sub_classes = [QuadraticBezierCurveElement]


class PathAParams(TypedDict):
    """
    Optional parameter definition for class PathA
    """

    angle: float | osvg.float.Float = (0,)
    long_bow: bool = (True,)
    right_bend: bool = (True,)


class PathA(_PathElement):
    """
    Class to draw a elliptic arc bow:
    - position: Final position of the arc curve
    - width: Width of the ellipse
    - height: Height of the ellipse
    - angle: Angled orientation of the ellipse
    - long_bow: Indicator if long bow should be drawn
    - right_bend:  Inficator if bow is right or left bend
    """

    width = osvg.float.FloatProperty()
    height = osvg.float.FloatProperty()
    angle = osvg.float.FloatProperty()

    def __init__(
        self,
        position: osvg.positions.Position,
        width: float | osvg.float.Float,
        height: float | osvg.float.Float,
        **kwargs: Unpack[PathAParams],
    ) -> None:
        self.position = position
        self.width = width
        self.height = height
        self.angle = kwargs.get("angle", 0)
        self.long_bow = kwargs.get("long_bow", True)
        self.right_bend = kwargs.get("right_bend", True)

    def __str__(self) -> str:
        return (
            f"A {self.width},{str(self.height)} "
            + f"{str(self.angle)},{int(self.long_bow)},{int(self.right_bend)} {str(self.position)}"
        )


class Path(osvg.elements.elementbase.SVGElement):
    """
    Class representing a SVG path element
    """

    xml_tag = "path"

    def __init__(
        self, **kwargs: Unpack[osvg.elements.elementbase.SVGElementParams]
    ) -> None:
        super().__init__(**kwargs)
        self.path_elements = [PathM(position=self.position)]

    def add_path_element(self, element: _PathElement) -> None:
        """
        Add a path element to the SVG path element
        """
        if element.needs_previous_path_element:
            element.verify_previous_path_element(self.path_elements[-1])
        self.path_elements.append(element)

    @property
    def _plain_etree_element(self) -> ET.Element:
        """
        Create ET for this object without wrappers
        """
        # Create the ET Element
        element = super()._plain_etree_element
        element.set("d", " ".join([str(x) for x in self.path_elements]))
        return element



"""
Module to define Group class.
"""

import osvg.elements.elementbase
import osvg.helper
import osvg.positions
import osvg.float
import osvg.style


class Group(osvg.elements.elementbase.SVGElement):
    """
    A Group class matches the 'g' tag in SVG.
    It can be customized with:
        - name => tag id
        - rotation: rotate group (all elements within it)
        - style => tag style
        - layer: Manipulate the order in the XML elements
    """

    xml_tag = "g"

    def add_group(
        self,
        position: "Group" = None,
        name: str = None,
        style: osvg.style.Style = None,
        layer: int = 0,
        rotation: float | osvg.float.Float = 0,
    ) -> "Group":
        """
        Add a group as a subgroup to this group.
        """
        # pylint: disable=too-many-arguments
        if position is None:
            position = osvg.positions.Position(x=self.position.x, y=self.position.y)
        return Group(
            parent=self,
            position=position,
            name=name,
            style=style,
            layer=layer,
            rotation=rotation,
        )
